# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

from ..base_model import BaseModel
from ..layers import EmbTransform, OneHotTransform
from ..scalers import TorchMinMaxScaler, TorchStandardScaler
from ..util import filter_nan

import GPy
import torch
import torch.nn as nn
import numpy as np

from torch import Tensor, FloatTensor, LongTensor

import logging
logging.disable(logging.WARNING)

import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)

class GPyGP(BaseModel):
    """
    Input warped GP model implemented using GPy instead of GPyTorch

    Why doing so:
    - Input warped GP
    """
    def __init__(self, num_cont, num_enum, num_out, **conf):
        super().__init__(num_cont, num_enum, num_out, **conf)
        if num_enum > 0:
            self.one_hot = OneHotTransform(self.conf['num_uniqs'])
        self.xscaler    = TorchMinMaxScaler((-1, 1))
        self.yscaler    = TorchStandardScaler()
        self.verbose    = self.conf.get('verbose', False)
        self.num_epochs = self.conf.get('num_epochs', 200)
        self.warp       = self.conf.get('warp', True)
        self.space      = self.conf.get('space') # DesignSpace
        if self.space is None and self.warp:
            warnings.warn('Space not provided, set warp to False')
            self.warp = False

    def fit_scaler(self, Xc : FloatTensor, y : FloatTensor):
        if Xc is not None and Xc.shape[1] > 0:
            if self.space is not None:
                cont_lb = self.space.opt_lb[:self.space.num_numeric].view(1, -1).float()
                cont_ub = self.space.opt_ub[:self.space.num_numeric].view(1, -1).float()
                self.xscaler.fit(torch.cat([Xc, cont_lb, cont_ub], dim = 0))
            else:
                self.xscaler.fit(Xc)
        self.yscaler.fit(y)

    def trans(self, Xc : Tensor, Xe : Tensor, y : Tensor = None):
        if Xc is not None and Xc.shape[1] > 0:
            Xc_t = self.xscaler.transform(Xc)
        else:
            Xc_t = torch.zeros(Xe.shape[0], 0)

        if Xe is None or Xe.shape[1] == 0:
            Xe_t = torch.zeros(Xc.shape[0], 0)
        else:
            Xe_t = self.one_hot(Xe.long())
        Xall = torch.cat([Xc_t, Xe_t], dim = 1)

        if y is not None:
            y_t = self.yscaler.transform(y)
            return Xall.numpy(), y_t.numpy()
        return Xall.numpy()

    def fit(self, Xc : FloatTensor, Xe : LongTensor, y : LongTensor): 
        Xc, Xe, y = filter_nan(Xc, Xe, y, 'all')
        self.fit_scaler(Xc, y)
        X, y = self.trans(Xc, Xe, y)

        k1  = GPy.kern.Linear(X.shape[1],   ARD = False)
        k2  = GPy.kern.Matern32(X.shape[1], ARD = True)
        k2.lengthscale = np.std(X, axis = 0)
        k2.variance    = 0.5
        k2.variance.set_prior(GPy.priors.Gamma(0.5, 1), warning = False)
        kern = k1 + k2
        if not self.warp:
            self.gp = GPy.models.GPRegression(X, y, kern)
        else:
            xmin    = np.zeros(X.shape[1])
            xmax    = np.ones(X.shape[1])
            xmin[:Xc.shape[1]] = -1
            warp_f  = GPy.util.input_warping_functions.KumarWarping(X, Xmin = xmin, Xmax = xmax)
            self.gp = GPy.models.InputWarpedGP(X, y, kern, warping_function = warp_f)
        self.gp.likelihood.variance.set_prior(GPy.priors.LogGaussian(-4.63, 0.5), warning = False)

        self.gp.optimize_restarts(max_iters = self.num_epochs, verbose = self.verbose, num_restarts = 10)
        return self

    def predict(self, Xc : FloatTensor, Xe : LongTensor) -> (FloatTensor, FloatTensor):
        Xall    = self.trans(Xc, Xe)
        py, ps2 = self.gp.predict(Xall)
        mu      = self.yscaler.inverse_transform(FloatTensor(py).view(-1, 1))
        var     = (self.yscaler.std**2 * FloatTensor(ps2).view(-1, 1)).clamp(min = 1e-6)
        return mu, var

    def sample_f(self):
        raise NotImplementedError('Thompson sampling is not supported for GP, use `sample_y` instead')

    @property
    def noise(self):
        var_normalized = self.gp.likelihood.variance[0]
        return (var_normalized * self.yscaler.std**2).view(self.num_out)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
import numpy as np
import torch
import torch.nn as nn
from hebo.models.util import filter_nan

def test_filter():
    x       = torch.randn(10, 1)
    y       = torch.randn(10, 3)
    y[0]    = np.nan
    y[1, 1] = np.nan

    xf, xef, yf = filter_nan(x, None, y)
    assert(yf.shape[0] == 9)

    xf, xef, yf = filter_nan(x, None, y, keep_rule = 'all')
    assert(yf.shape[0] == 8)
    assert(torch.isfinite(yf).all())



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy as np
import pandas as pd
from sklearn.model_selection import cross_val_predict, KFold
from sklearn.metrics import get_scorer, make_scorer
from typing import Callable

from hebo.design_space.design_space import DesignSpace
from hebo.optimizers.hebo import HEBO

import warnings
warnings.filterwarnings('ignore')

def sklearn_tuner(
        model_class, 
        space_config : [dict], 
        X : np.ndarray,
        y : np.ndarray,
        metric : Callable, 
        greater_is_better : bool = True, 
        n_splits = 5,
        max_iter = 16, 
        report   = False
        ) -> (dict, pd.DataFrame):
    """Tuning sklearn estimator

    Parameters:
    -------------------
    model_class: class of sklearn estimator
    space_config: list of dict, specifying search space
    X, y: data used to for cross-valiation
    metrics: metric function in sklearn.metrics
    greater_is_better: whether a larger metric value is better
    n_splits: split data into `n_splits` parts for cross validation
    max_iter: number of trials

    Returns:
    -------------------
    Best hyper-parameters and all visited data


    Example:
    -------------------
    from sklearn.datasets import load_boston
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.metrics import r2_score, mean_squared_error
    from hebo.sklearn_tuner import sklearn_tuner

    space_cfg = [
            {'name' : 'max_depth',        'type' : 'int', 'lb' : 1, 'ub' : 20},
            {'name' : 'min_samples_leaf', 'type' : 'num', 'lb' : 1e-4, 'ub' : 0.5},
            {'name' : 'max_features',     'type' : 'cat', 'categories' : ['auto', 'sqrt', 'log2']},
            {'name' : 'bootstrap',        'type' : 'bool'},
            {'name' : 'min_impurity_decrease', 'type' : 'pow', 'lb' : 1e-4, 'ub' : 1.0},
            ]
    X, y   = load_boston(return_X_y = True)
    result = sklearn_tuner(RandomForestRegressor, space_cfg, X, y, metric = r2_score, max_iter = 16)
    """
    space = DesignSpace().parse(space_config)
    opt   = HEBO(space)
    for i in range(max_iter):
        rec     = opt.suggest()
        model   = model_class(**rec.iloc[0].to_dict())
        pred    = cross_val_predict(model, X, y, cv = KFold(n_splits = n_splits, shuffle = True))
        score_v = metric(y, pred)
        sign    = -1. if greater_is_better else 1.0
        opt.observe(rec, np.array([sign * score_v]))
        print('Iter %d, best metric: %g' % (i, sign * opt.y.min()))
    best_id   = np.argmin(opt.y.reshape(-1))
    best_hyp  = opt.X.iloc[best_id]
    df_report = opt.X.copy()
    df_report['metric'] = sign * opt.y
    if report:
        return best_hyp.to_dict(), df_report
    else:
        return best_hyp.to_dict()



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

from .base_model import BaseModel
from .gp.gp import GP
from .gp.gpy_wgp import GPyGP
from .gp.gpy_mlp import GPyMLPGP
from .rf.rf import RF
from .nn.deep_ensemble import DeepEnsemble

model_dict = {
        'gp'  : GP,
        'gpy' : GPyGP,
        'gpy_mlp' : GPyMLPGP, 
        'rf'  : RF,
        'deep_ensemble' : DeepEnsemble
        }

def get_model(model_name : str, *params, **conf) -> BaseModel:
    assert model_name in model_dict, "model name %s not in model_dict"
    model_class = model_dict[model_name]
    return model_class(*params, **conf)

def get_model_class(model_name : str, *params, **conf):
    assert model_name in model_dict, "model name %s not in model_dict"
    model_class = model_dict[model_name]
    return model_class



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')

from hebo.acq_optimizers.evolution_optimizer import EvolutionOpt
from hebo.acquisitions.acq import  Acquisition 
from hebo.design_space.design_space import DesignSpace 

import pytest
from pytest import approx

import torch

class ToyExample(Acquisition):
    def __init__(self, constr_v = 1.0):
        super().__init__(None)
        self.constr_v = constr_v
    
    @property
    def num_obj(self):
        return 1

    @property
    def num_constr(self):
        return 1

    def eval(self, x, xe):
        # minimize L2norm(x) s.t. L2norm(x) > constr_v
        out    = (x**2).sum(dim = 1).view(-1, 1)
        constr = self.constr_v - out  
        return torch.cat([out, constr], dim = 1)

class ToyExampleMO(Acquisition):
    def __init__(self):
        super().__init__(None)
    
    @property
    def num_obj(self):
        return 2

    @property
    def num_constr(self):
        return 0

    def eval(self, x, xe):
        # minimize L2norm(x) s.t. L2norm(x) > 1.0
        o1 = (x**2).sum(dim = 1).view(-1, 1)
        o2 = ((x-1)**2).sum(dim = 1).view(-1, 1)
        return torch.cat([o1, o2], dim = 1)

def test_opt():
    space = DesignSpace().parse([
        {'name' : 'x1', 'type' : 'num', 'lb' : -3.0, 'ub' : 3.0}
        ])
    acq   = ToyExample()
    opt   = EvolutionOpt(space, acq, pop = 10)
    rec   = opt.optimize(initial_suggest = space.sample(3))
    x, xe = space.transform(rec)
    assert(approx(1.0, 1e-3) == acq(x, xe)[:, 0].squeeze().item())

def test_opt_fix():
    space = DesignSpace().parse([
        {'name' : 'x1', 'type' : 'num', 'lb' : -3.0, 'ub' : 3.0}, 
        {'name' : 'x2', 'type' : 'num', 'lb' : -3.0, 'ub' : 3.0}
        ])
    acq   = ToyExample()
    opt   = EvolutionOpt(space, acq, pop = 10)
    rec   = opt.optimize(fix_input = {'x1' : 1.0})
    print(rec)
    assert (rec['x1'].values == approx(1.0, 1e-3))

def test_opt_int():
    space = DesignSpace().parse([
        {'name' : 'x1', 'type' : 'num', 'lb' : -3.0, 'ub' : 3.0}, 
        {'name' : 'x2', 'type' : 'int', 'lb' : -3.0, 'ub' : 3.0}
        ])
    acq   = ToyExample()
    opt   = EvolutionOpt(space, acq, pop = 10)
    rec   = opt.optimize()
    assert(approx(1.0, 1e-3) == acq(*space.transform(rec))[:, 0].squeeze().item())

def test_mo():
    space = DesignSpace().parse([
        {'name' : 'x1', 'type' : 'num', 'lb' : -3.0, 'ub' : 3.0}, 
        {'name' : 'x2', 'type' : 'int', 'lb' : -3.0, 'ub' : 3.0}
        ])
    acq = ToyExampleMO()
    opt = EvolutionOpt(space, acq, pop = 10)
    rec = opt.optimize()
    assert(rec.shape[0] == 10)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')

import pytest
from pytest import approx

from hebo.design_space.design_space      import DesignSpace
from hebo.design_space.numeric_param     import NumericPara
from hebo.design_space.integer_param     import IntegerPara
from hebo.design_space.pow_param         import PowPara
from hebo.design_space.categorical_param import CategoricalPara
from hebo.design_space.bool_param        import BoolPara

def test_design_space():
    space = DesignSpace().parse([
        {'name' : 'x0', 'type' : 'num', 'lb' : 0, 'ub' : 7}, 
        {'name' : 'x1', 'type' : 'int', 'lb' : 0, 'ub' : 7}, 
        {'name' : 'x2', 'type' : 'pow', 'lb' : 1e-4, 'ub' : 1e-2, 'base' : 10}, 
        {'name' : 'x3', 'type' : 'cat', 'categories' : ['a', 'b', 'c']}, 
        {'name' : 'x4', 'type' : 'bool'}
    ])
    assert space.numeric_names   == ['x0', 'x1', 'x2', 'x4']
    assert space.enum_names      == ['x3']
    assert space.num_paras       == 5
    assert space.num_numeric     == 4
    assert space.num_categorical == 1

    samp    = space.sample(10)
    x, xe   = space.transform(samp)
    x_, xe_ = space.transform(space.inverse_transform(x, xe))
    assert (x - x_).abs().max() < 1e-4

    assert (space.opt_lb <= space.opt_ub).all()

    assert not space.paras['x0'].is_discrete
    assert space.paras['x1'].is_discrete
    assert not space.paras['x2'].is_discrete
    assert space.paras['x3'].is_discrete
    assert space.paras['x4'].is_discrete



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
import torch

from pytest import approx
from hebo.models.gp.gp import GP

def test_noise_free():
    X = torch.randn(10, 3)
    y = (X**2).sum(dim = 1).view(-1, 1) + 0.01 * torch.randn(10, 1)
    model = GP(3, 0, 1, noise_free = True, num_epochs = 10)
    model.fit(X, None, y)
    assert(approx(model.noise.item()) == 1.1 * model.noise_lb * model.yscaler.std**2)

@pytest.mark.xfail
def test_pred_likeli():
    X = torch.randn(10, 3)
    y = (X**2).sum(dim = 1).view(-1, 1) + 0.01 * torch.randn(10, 1)
    model = GP(3, 0, 1, noise_free = True, num_epochs = 10, pred_likeli = False)
    model.fit(X, None, y)

    with torch.no_grad():
        py, ps2 = model.predict(X, None)
        samp    = model.sample_y(X, None)

        model.pred_likeli = False

        py_no_noise, ps2_no_noise = model.predict(X, None)
        samp_no_noise = model.sample_y(X, None)
        assert((py_no_noise == py).all())

        delta = ps2 - ps2_no_noise
        assert((delta > 0).all())

        assert((samp != samp_no_noise).all())

def test_verbose(capsys):
    X = torch.randn(10, 3)
    y = (X**2).sum(dim = 1).view(-1, 1) + 0.01 * torch.randn(10, 1)
    model = GP(3, 0, 1, num_epochs = 10, verbose = True)
    model.fit(X, None, y)

    out, err = capsys.readouterr()
    assert('After'  in out)
    assert('epochs' in out)
    assert('loss'   in out)
    assert(err == '')

    model = GP(3, 0, 1, num_epochs = 10, verbose = False)
    model.fit(X, None, y)

    out, err = capsys.readouterr()
    assert(out == '')
    assert(err == '')



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import pandas as pd
import numpy as np
from abc import ABC, abstractmethod

from hebo.design_space.design_space import DesignSpace

class AbstractOptimizer(ABC):
    support_parallel_opt    = False
    support_constraint      = False
    support_multi_objective = False
    support_combinatorial   = False
    support_contextual      = False
    def __init__(self, space : DesignSpace):
        self.space = space

    def suggest(self, n_suggestions = 1, fix_input : dict = None):
        """
        Perform optimisation and give recommendation using data observed so far
        ---------------------
        n_suggestions:  number of recommendations in this iteration

        fix_input:      parameters NOT to be optimized, but rather fixed, this
                        can be used for contextual BO, where you can set the contex as a design
                        parameter and fix it during optimisation
        """
        pass

    def observe(self, x : pd.DataFrame, y : np.ndarray):
        """
        Observe new data
        """
        pass



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy as np
from sklearn.preprocessing import LabelEncoder
from .param import Parameter

class BoolPara(Parameter):
    def __init__(self, param):
        super().__init__(param)
        self.lb         = 0
        self.ub         = 1

    def sample(self, num = 1):
        assert(num > 0)
        return np.random.choice([True, False], num, replace = True)

    def transform(self, x):
        return x.astype(float)

    def inverse_transform(self, x):
        return x > 0.5

    @property
    def is_numeric(self):
        # XXX: It's OK to view boolean as numeric value, this may reduce
        # dimensions if catecorical variables are procecessed via one-hot or
        # embedding
        return True 

    @property
    def is_discrete(self):
        return True

    @property
    def opt_lb(self):
        return self.lb

    @property
    def opt_ub(self):
        return self.ub



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
from pytest import approx

import torch
import numpy as np
from torch import FloatTensor
from sklearn.metrics import r2_score

from hebo.models.base_model import BaseModel
from hebo.models.model_factory import get_model, model_dict

def rand_dataset(num_out : int) -> (FloatTensor, FloatTensor):
    X = torch.randn(10, 2)
    y = torch.zeros(10, num_out)
    for i in range(num_out):
        w = torch.randn(2)
        y[:, i] = (X * w).sum(axis = 1)
    return X, y

def rand_dataset_with_enum(num_out : int) -> (FloatTensor, FloatTensor):
    X  = torch.randn(10, 2)
    Xe = torch.randint(2, (10, 2))
    y  = torch.zeros(10, num_out)
    for i in range(num_out):
        w = torch.randn(2)
        y[:, i] = (X * w).sum(axis = 1)
    return X, Xe, y

def gather_mo_models():
    model_list = []
    for name in model_dict:
        try:
            model = get_model(name, 2, 0, 2)
            model_list.append(name)
        except AssertionError as e:
            assert str(e) == "Model only support single-output"
    return model_list       

@pytest.fixture(params = gather_mo_models())
def model_name(request):
    return request.param

@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_init(model_name, num_out):
    model = get_model(model_name, 2, 0, num_out)
    assert(model.num_out == num_out)

@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_fit(model_name, num_out):
    """
    Model should be able to overfit simple linear model with training R2 > 0.8
    """
    model = get_model(model_name, 2, 0, num_out, num_epochs = 20)
    X, y  = rand_dataset(num_out)
    model.fit(X, None, y)
    with torch.no_grad():
        py, ps2 = model.predict(X, None)
        ps      = ps2.sqrt()
        assert(r2_score(y.numpy(), py.numpy()) > 0.8)
        assert (py + 3 * ps > y).sum() > 0.9 * y.numel()
        assert (py - 3 * ps < y).sum() > 0.9 * y.numel()

@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_fit_with_enum(model_name, num_out):
    """
    Model should be able to overfit simple linear model with training R2 > 0.8
    """
    model    = get_model(model_name, 2, 2, num_out, num_uniqs = [2, 2], num_epochs = 20)
    X, Xe, y = rand_dataset_with_enum(num_out)
    model.fit(X, Xe, y)
    with torch.no_grad():
        py, ps2 = model.predict(X, Xe)
        ps      = ps2.sqrt()
        assert(r2_score(y.numpy(), py.numpy()) > 0.8)
        assert (py + 3 * ps > y).sum() > 0.9 * y.numel()
        assert (py - 3 * ps < y).sum() > 0.9 * y.numel()

@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_grad(model_name, num_out):
    model = get_model(model_name, 2, 0, num_out, num_epochs = 2)
    X, y  = rand_dataset(num_out)
    model.fit(X, None, y)
    X.requires_grad = True
    py, ps2 = model.predict(X, None)
    ps      = ps2.sqrt()
    y = (py + ps).sum()
    y.backward()
    assert(X.grad.shape == X.shape)

    delta = 0.001
    Xdelta = X.detach().clone()
    Xdelta[0, 0] += delta
    py, ps2 = model.predict(Xdelta, None)
    ydelta  = (py + ps2.sqrt()).sum()
    grad_delta = ((ydelta - y) / delta).detach()
    assert(grad_delta.numpy() == approx(X.grad[0, 0].numpy(), abs = 0.2))

@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_sample(model_name, num_out):
    model = get_model(model_name, 2, 0, num_out, num_epochs = 1)
    X, y = rand_dataset(num_out)
    model.fit(X, None, y)
    assert(model.sample_y(X, None, 5).shape == torch.Size([5, X.shape[0], y.shape[1]]))
    
@pytest.mark.parametrize('num_out', [1, 2], ids = ['single-out', 'multi-out'])
def test_noise(model_name, num_out):
    model = get_model(model_name, 2, 0, num_out, num_epochs = 5)
    X, y = rand_dataset(num_out)
    model.fit(X, None, y)
    assert(model.noise.shape == torch.Size([model.num_out]))
    assert((model.noise >= 0).all())



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
from pytest import approx

import torch
import numpy as np
from torch import FloatTensor
from sklearn.metrics import r2_score

from hebo.models.base_model import BaseModel
from hebo.models.model_factory import get_model, get_model_class, model_dict

def check_overfitted(y_true : FloatTensor, py_pred : FloatTensor, ps2_pred : FloatTensor) -> bool:
    y  = y_true.numpy().reshape(-1)
    py = py_pred.numpy().reshape(-1)
    ps = ps2_pred.sqrt().numpy().reshape(-1)
    assert r2_score(y, py) > 0.7
    assert (py + 3 * ps >= y).sum() > 0.9 * y.size
    assert (py - 3 * ps <= y).sum() > 0.9 * y.size

@pytest.fixture(params = list(model_dict.keys()))
def model_name(request):
    return request.param

def test_is_basic_model(model_name):
    assert(issubclass(type(get_model(model_name, 1, 0, 1)), BaseModel))

def test_model_can_be_initialized(model_name):
    model = get_model(model_name, 1, 0, 1)
    model_class = get_model_class(model_name)
    model = model_class(1, 0, 1)

def test_fit_with_cont_enum(model_name):
    Xc = torch.randn(50, 1)
    Xe = torch.randint(2, (50, 1))
    y  = torch.zeros(50, 1)
    y[Xe.squeeze() == 1] = Xc[Xe.squeeze() == 1]
    y[Xe.squeeze() == 0] = -1 * Xc[Xe.squeeze() == 0]
    model = get_model(model_name, 1, 1, 1, num_uniqs = [2], num_epochs = 100)
    model.fit(Xc, Xe, y + 1e-2 * torch.randn(y.shape))
    with torch.no_grad():
        py, ps2 = model.predict(Xc, Xe)
        check_overfitted(y, py, ps2)
        samp = model.sample_y(Xc, Xe, 50).mean(axis = 0)
        assert r2_score(y.numpy(), samp.numpy()) > 0.8

def test_fit_with_cont_only(model_name):
    Xc = torch.randn(50, 1)
    Xe = None
    y  = Xc
    model = get_model(model_name, 1, 0, 1, num_epochs = 30)
    model.fit(Xc, Xe, y + 1e-2 * torch.randn(y.shape))
    with torch.no_grad():
        py, ps2 = model.predict(Xc, Xe)
        check_overfitted(y, py, ps2)

def test_fit_with_enum_only(model_name):
    Xc = None
    Xe = torch.randint(2, (50, 1))
    y  = Xe.float()
    model = get_model(model_name, 0, 1, 1, num_uniqs = [2], num_epochs = 30)
    model.fit(Xc, Xe, y + 1e-2 * torch.randn(y.shape))
    with torch.no_grad():
        py, ps2 = model.predict(Xc, Xe)
        check_overfitted(y, py, ps2)

def test_thompson_sampling(model_name):
    model = get_model(model_name, 1, 0, 1)
    Xc = torch.randn(10, 1)
    Xe = None
    y  = Xc ** 2
    model = get_model(model_name, 1, 0, 1, num_epochs = 30)
    model.fit(Xc, Xe, y + 1e-2 * torch.randn(y.shape))

    with torch.no_grad():
        if model.support_ts:
            f  = model.sample_f()
            py = f(Xc, Xe)
            assert(r2_score(y.numpy(), py.numpy()) > 0.7)
        else:
            with pytest.raises(NotImplementedError):
                f = model.sample_f()

def test_grad(model_name):
    Xc = torch.randn(50, 1)
    Xe = None
    y  = Xc + 0.01 * torch.randn(50, 1)
    model = get_model(model_name, 1, 0, 1, num_epochs = 30)
    model.fit(Xc, Xe, y)
    if model.support_grad:
        X_tst = torch.randn(50, 1)
        X_tst.requires_grad = True
        py, _ = model.predict(X_tst, None)
        obj = py.sum()
        obj.backward()
        assert(X_tst.grad is not None)
    else:
        pytest.skip('%s does not support grad' % model_name)

def test_noise(model_name):
    Xc = torch.randn(50, 1)
    Xe = None
    y  = Xc + 0.01 * torch.randn(50, 1)
    model = get_model(model_name, 1, 0, 1, num_epochs = 1)
    model.fit(Xc, Xe, y)
    with torch.no_grad():
        assert(model.noise.shape == torch.Size([model.num_out]))
        assert(super(type(model), model).noise.shape == torch.Size([model.num_out]))
        assert((model.noise >= 0).all())

def test_warm_start(model_name):
    Xc    = torch.randn(64, 1)
    y     = Xc + 0.01 * torch.randn(Xc.shape[0], 1)
    model = get_model(model_name, 1, 0, 1, num_epochs = 5, output_noise = False, batch_size = 64)
    if model.support_warm_start:
        model.fit(Xc, None, y)
        with torch.no_grad():
            py, ps2 = model.predict(Xc, None)
        err1 = ((py - y)**2).mean()
        model.fit(Xc, None, y)
        with torch.no_grad():
            py, ps2 = model.predict(Xc, None)
        err2 = ((py - y)**2).mean()
        assert(err2 < err1)
    else:
        pytest.skip('%s does not support warm start' % model_name)

def test_fit_with_nan(model_name):
    x     = torch.randn(10, 1)
    y     = x + 1e-4 * torch.randn(10, 1)
    y[0]  = np.nan
    model = get_model(model_name, 1, 0, 1, num_epochs = 5)
    try:
        model.fit(x, None, y)
    except:
        assert False, "Model won't fit with NaN"
    with torch.no_grad():
        py, ps2 = model.predict(x, None)
        assert(torch.isfinite(py).all())
        assert(torch.isfinite(ps2).all())



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
import torch

from hebo.models.nn.deep_ensemble import BaseNet, DeepEnsemble
from sklearn.metrics import r2_score

@pytest.mark.parametrize('num_cont',  [0, 1])
@pytest.mark.parametrize('num_enum', [0, 1])
@pytest.mark.parametrize('num_out',   [1, 2])
@pytest.mark.parametrize('output_noise', [True, False])
@pytest.mark.parametrize('rand_prior', [True, False])
@pytest.mark.parametrize('num_layers', [1, 2])
def test_basenet(num_cont, num_enum, num_out, output_noise, rand_prior, num_layers):
    if num_cont + num_enum > 0:
        xc  = torch.randn(10, num_cont)
        xe  = torch.randint(2, (10, num_enum))
        y   = torch.randn(10, num_out)
        net = BaseNet(num_cont, num_enum, num_out, num_uniqs = [2], output_noise = output_noise, rand_prior = rand_prior, num_layers = num_layers)
        py  = net(xc, xe)
        assert(torch.isfinite(py).all())
        assert(py.shape[0] == y.shape[0])
        assert(py.shape[1] == y.shape[1] if not output_noise else y.shape[1] * 2) 
        if output_noise:
            assert ((py[:, -1 * net.num_out:] >= 0)).all()
        
        crit = torch.nn.MSELoss()
        mse1 = crit(y, py[:, :net.num_out]).item()
        
        opt = torch.optim.Adam(net.parameters(), lr = 1e-3)
        for i in range(5):
            py   = net(xc, xe)
            loss = crit(y, py[:, :net.num_out])
            opt.zero_grad()
            loss.backward()
            opt.step()
        mse2 = loss.item()
        assert(mse2 < mse1)

@pytest.mark.parametrize('output_noise', [True, False], ids = ['nll', 'mse'])
@pytest.mark.parametrize('rand_prior',   [True, False], ids = ['with_prior', 'no_prior'])
@pytest.mark.parametrize('num_processes', [1, 5], ids = ['seq', 'para'])
def test_deep_ens(output_noise, rand_prior, num_processes):
    xc = torch.randn(16, 1)
    y  = xc ** 2
    model = DeepEnsemble(1, 0, 1, 
            output_noise  = output_noise,
            rand_prior    = rand_prior,
            num_processes = num_processes,
            l1            = 0., 
            num_epoch     = 16)
    model.fit(xc, None, y)
    with torch.no_grad():
        py, _ = model.predict(xc, None)
        assert(r2_score(y.numpy(), py.numpy()) > 0.8)

def test_verbose(capsys):
    X = torch.randn(10, 3)
    y = (X**2).sum(dim = 1).view(-1, 1) + 0.01 * torch.randn(10, 1)
    model = DeepEnsemble(3, 0, 1, num_epochs = 10, verbose = True)
    model.fit(X, None, y)

    out, err = capsys.readouterr()
    assert('Epoch'  in out)
    assert('loss'   in out)
    assert(err == '')

    model = DeepEnsemble(3, 0, 1, num_epochs = 10, verbose = False)
    model.fit(X, None, y)

    out, err = capsys.readouterr()
    assert(out == '')
    assert(err == '')



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

from hebo.design_space.design_space import DesignSpace
def parse_space_from_bayesmark(api_config) -> DesignSpace:
    """
    Parse design space of bayesmark (https://github.com/uber/bayesmark)
    """
    space  = DesignSpace()
    params = []
    for param_name in api_config:
        param_conf   = api_config[param_name]
        param_type   = param_conf['type']
        param_space  = param_conf.get('space', None)
        param_range  = param_conf.get("range", None)
        param_values = param_conf.get("values", None)

        bo_param_conf = {'name' : param_name}
        if param_type == 'int': # ignore 'log' space # TODO: support log-scale int
            bo_param_conf['type'] = 'int'
            bo_param_conf['lb']   = param_range[0]
            bo_param_conf['ub']   = param_range[1]
        elif param_type == 'bool':
            bo_param_conf['type'] = 'bool'
        elif param_type in ('cat', 'ordinal'):
            bo_param_conf['type']       = 'cat'
            bo_param_conf['categories'] = list(set(param_values))
        elif param_type == 'real':
            if param_space in ('log', 'logit'):
                bo_param_conf['type'] = 'pow'
                bo_param_conf['base'] = 10
                bo_param_conf['lb']   = param_range[0]
                bo_param_conf['ub']   = param_range[1]
            else:
                bo_param_conf['type'] = 'num'
                bo_param_conf['lb']   = param_range[0]
                bo_param_conf['ub']   = param_range[1]
        else:
            assert False, "type %s not handled in API" % param_type
        params.append(bo_param_conf)
    space.parse(params)
    return space



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import os
import sys

import traceback
import numpy as np
import torch
import torch.nn as nn

from multiprocessing import Pool
from pathlib import Path
from torch.utils.data import DataLoader, TensorDataset
from torch import Tensor, FloatTensor, LongTensor
from copy import deepcopy

from ..base_model import BaseModel
from ..layers import EmbTransform
from ..scalers import TorchMinMaxScaler, TorchStandardScaler

class DeepEnsemble(BaseModel):
    support_ts           = True
    support_grad         = True
    support_multi_output = True
    support_warm_start   = True
    def __init__(self, num_cont, num_enum, num_out, **conf):
        super().__init__(num_cont, num_enum, num_out, **conf)
        self.bootstrap     = conf.get('bootstrap',     False)
        self.rand_prior    = conf.get('rand_prior',    False)
        self.output_noise  = conf.get('output_noise',  True)
        self.num_ensembles = conf.get('num_ensembles', 5)
        self.num_process   = conf.get('num_processes', 1)
        self.num_epochs    = conf.get('num_epochs',    500)
        self.print_every   = conf.get('print_every',   50)

        self.num_layers    = conf.get('num_layers',    1)
        self.num_hiddens   = conf.get('num_hiddens',   128)
        self.l1            = conf.get('l1',            1e-3)
        self.batch_size    = conf.get('batch_size',    32)
        self.lr            = conf.get('lr',            5e-3)
        self.adv_eps       = conf.get('adv_eps',       0.)
        self.verbose       = conf.get('verbose',       False)
        assert self.num_ensembles > 0

        self.xscaler = TorchMinMaxScaler((-1, 1))
        self.yscaler = TorchStandardScaler()

        self.loss       = self.loss_likelihood if self.output_noise else self.loss_mse
        self.loss_name  = "NLL" if self.output_noise else "MSE"
        self.models     = None
        self.sample_idx = 0
        self.noise_est  = torch.zeros(self.num_out)

    @property
    def fitted(self):
        return self.models is not None

    @property
    def noise(self) -> FloatTensor:
        return self.noise_est

    def fit(self, Xc_ : FloatTensor, Xe_ : LongTensor, y_ : FloatTensor):
        valid = torch.isfinite(y_).any(dim = 1)
        Xc    = Xc_[valid] if Xc_ is not None else None
        Xe    = Xe_[valid] if Xe_ is not None else None
        y     = y_[valid]

        if self.models is None:
            self.fit_scaler(Xc, Xe, y)
        Xc, Xe, y = self.trans(Xc, Xe, y)

        if self.num_process > 1:
            with Pool(self.num_process) as p:
                self.models = p.starmap(self.fit_one, [(Xc.clone(), Xe.clone(), y.clone(), model_idx) for model_idx in range(self.num_ensembles)]) 
        else:
            self.models = [self.fit_one(Xc, Xe, y, i) for i in range(self.num_ensembles)]
    
        assert None not in self.models
        self.sample_idx = 0

        with torch.no_grad():
            py, _ = self.predict(Xc_, Xe_)
            err   = (py - y_)[valid]
            self.noise_est = (err**2).mean(dim = 0).detach().clone()

    def predict(self, Xc_ : FloatTensor, Xe_ : LongTensor) -> (FloatTensor, FloatTensor):
        Xc, Xe = self.trans(Xc_, Xe_)
        preds  = torch.stack([self.models[i](Xc, Xe) for i in range(self.num_ensembles)])
        if not self.output_noise:
            py  = preds.mean(dim = 0)
            ps2 = 1e-8 + preds.var(dim = 0, unbiased = False) # XXX: var([1.0], unbiased = True) = NaN
        else:
            mu     = preds[:, :, :self.num_out]
            sigma2 = preds[:, :, self.num_out:]
            py     = mu.mean(dim = 0)
            ps2    = mu.var(dim = 0, unbiased = False) + sigma2.mean(axis = 0)
        return self.yscaler.inverse_transform(py), ps2 * self.yscaler.std**2

    def sample_f(self):
        assert(self.fitted)
        idx = self.sample_idx
        self.sample_idx = (self.sample_idx + 1) % self.num_ensembles
        def f(Xc : FloatTensor, Xe : LongTensor) -> FloatTensor:
            model  = self.models[idx]
            Xc, Xe = self.trans(Xc, Xe)
            return self.yscaler.inverse_transform(model(Xc, Xe)[:, :self.num_out])
        return f

    def fit_scaler(self, Xc : Tensor, Xe : Tensor, y : Tensor):
        if Xc is not None and Xc.shape[1] > 0:
            self.xscaler.fit(Xc)
        self.yscaler.fit(y)
    
    def trans(self, Xc : Tensor, Xe : Tensor, y : Tensor = None):
        if Xc is not None and Xc.shape[1] > 0:
            Xc_t = self.xscaler.transform(Xc)
        else:
            Xc_t = torch.zeros(Xe.shape[0], 0)

        if Xe is None:
            Xe_t = torch.zeros(Xc.shape[0], 0).long()
        else:
            Xe_t = Xe.long()

        if y is not None:
            y_t = self.yscaler.transform(y)
            return Xc_t, Xe_t, y_t
        else:
            return Xc_t, Xe_t

    def loss_mse(self, pred, target):
        mask = torch.isfinite(target)
        return nn.MSELoss()(pred[mask], target[mask])

    def loss_likelihood(self, pred, target):
        mask   = torch.isfinite(target)
        mu     = pred[:, :self.num_out][mask]
        sigma2 = pred[:, self.num_out:][mask]
        loss   = 0.5 * (target[mask] - mu)**2 / sigma2 + 0.5 * torch.log(sigma2)
        return torch.mean(loss)

    def fit_one(self, Xc, Xe, y, idx):
        torch.seed()
        dataset   = TensorDataset(Xc, Xe, y)
        loader    = DataLoader(dataset, batch_size = self.batch_size, shuffle = True)
        if self.models is not None and len(self.models) == self.num_ensembles:
            model = deepcopy(self.models[idx])
        else:
            model = BaseNet(self.num_cont, self.num_enum, self.num_out,
                    noise_lb     = 1e-4,
                    num_uniqs    = None if self.num_enum == 0 else self.conf['num_uniqs'], 
                    num_layers   = self.num_layers,
                    num_hiddens  = self.num_hiddens,
                    output_noise = self.output_noise,
                    rand_prior   = self.rand_prior)
        opt       = torch.optim.Adam(model.parameters(), lr = self.lr)
        model.train()
        for epoch in range(self.num_epochs):
            epoch_loss = 0
            for bxc, bxe, by in loader:
                py        = model(bxc, bxe)
                data_loss = self.loss(py, by)
                reg_loss  = 0.
                for p in model.parameters():
                    reg_loss += self.l1 * p.abs().sum() / (y.shape[0] * y.shape[1])
                loss = data_loss + reg_loss
                opt.zero_grad()
                loss.backward()
                opt.step()

                # XXX: Adversarial training not used

                epoch_loss += data_loss * bxc.shape[0]
            if epoch % self.print_every == 0:
                if self.verbose:
                    print("Epoch %d, %s loss = %g" % (epoch, self.loss_name, epoch_loss / Xc.shape[0]), flush = True)
        return model

class BaseNet(nn.Module):
    def __init__(self, num_cont, num_enum, num_out, **conf):
        super().__init__()
        self.num_cont     = num_cont
        self.num_enum     = num_enum
        self.num_out      = num_out
        self.noise_lb     = conf.get('noise_lb', 1e-4)
        self.num_layers   = conf.get('num_layers', 1)  # number of hidden layers
        self.num_hiddens  = conf.get('num_hiddens', 128) # number of hidden units
        self.output_noise = conf.get('output_noise', True)
        self.rand_prior   = conf.get('rand_prior', False)
        self.eff_dim      = num_cont
        if self.num_enum > 0:
            assert 'num_uniqs' in conf, "num_uniqs not in algorithm configuration"
            self.emb_trans  = EmbTransform(conf['num_uniqs'])
            self.eff_dim   += self.emb_trans.num_out

        self.hidden = self.construct_hidden()
        self.mu     = nn.Linear(self.num_hiddens, self.num_out)
        if self.output_noise:
            self.sigma2 = nn.Sequential(
                nn.Linear(self.num_hiddens, self.num_out), 
                nn.Softplus())
        if self.rand_prior: # Randomized Prior Functions for Deep Reinforcement Learning, Ian Osband
            self.prior_net = self.construct_hidden()
            self.prior_net.add_module('prior_net_out', nn.Linear(self.num_hiddens, self.num_out))
        for n, p in self.named_parameters():
            if "bias" in n:
                nn.init.zeros_(p)
            else:
                nn.init.xavier_uniform_(p, gain = nn.init.calculate_gain('relu'))

    def construct_hidden(self):
        layers = [nn.Linear(self.eff_dim, self.num_hiddens), nn.ReLU()]
        for i in range(self.num_layers - 1):
            layers.append(nn.Linear(self.num_hiddens, self.num_hiddens))
            layers.append(nn.ReLU())
        return nn.Sequential(*layers)

    def xtrans(self, Xc : FloatTensor, Xe : LongTensor) -> FloatTensor:
        Xall     = Xc.clone() if self.num_cont > 0 else torch.zeros(Xe.shape[0], 0)
        if self.num_enum > 0:
            Xall = torch.cat([Xall, self.emb_trans(Xe)], dim = 1)
        return Xall

    def forward(self, Xc : FloatTensor, Xe : LongTensor) -> FloatTensor:
        inputs = self.xtrans(Xc, Xe)
        prior_out = 0.
        if self.rand_prior:
            with torch.no_grad():
                prior_out = self.prior_net(inputs).detach()
        hidden = self.hidden(inputs)
        mu     = self.mu(hidden) + prior_out
        out    = torch.cat((mu, self.noise_lb + self.sigma2(hidden)), dim = 1) if self.output_noise else mu
        return out



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
import numpy as np
from torch import Tensor
from torch.distributions import Normal
from abc import ABC, abstractmethod
from ..models.base_model import BaseModel

class Acquisition(ABC):
    def __init__(self, model, **conf):
        self.model = model

    @property
    @abstractmethod
    def num_obj(self):
        pass

    @property
    @abstractmethod
    def num_constr(self):
        pass

    @abstractmethod
    def eval(self, x : Tensor,  xe : Tensor) -> Tensor:
        """
        Shape of output tensor: (x.shape[0], self.num_obj + self.num_constr)
        """
        pass

    def __call__(self, x : Tensor,  xe : Tensor):
        return self.eval(x, xe)

class SingleObjectiveAcq(Acquisition):
    """
    Single-objective, unconstrained acquisition
    """
    def __init__(self, model : BaseModel, **conf):
        super().__init__(model, **conf)

    @property
    def num_obj(self):
        return 1

    @property
    def num_constr(self):
        return 0

class LCB(SingleObjectiveAcq):
    def __init__(self, model : BaseModel, kappa = 3.0, **conf):
        super().__init__(model, **conf)
        self.kappa = kappa
        assert(model.num_out == 1)
    
    def eval(self, x : Tensor, xe : Tensor) -> Tensor:
        py, ps2 = self.model.predict(x, xe)
        return py - self.kappa * ps2.sqrt()

class Mean(SingleObjectiveAcq):
    def __init__(self, model : BaseModel, **conf):
        super().__init__(model, **conf)
        assert(model.num_out == 1)

    def eval(self, x : Tensor, xe : Tensor) -> Tensor:
        py, _ = self.model.predict(x, xe)
        return py

class Sigma(SingleObjectiveAcq):
    def __init__(self, model : BaseModel, **conf):
        super().__init__(model, **conf)
        assert(model.num_out == 1)

    def eval(self, x : Tensor, xe : Tensor) -> Tensor:
        _, ps2 = self.model.predict(x, xe)
        return -1 * ps2.sqrt()

class EI(SingleObjectiveAcq):
    pass

class logEI(SingleObjectiveAcq):
    pass

class WEI(Acquisition):
    pass

class Log_WEI(Acquisition):
    pass

class MES(SingleObjectiveAcq):
    pass

class MOMeanSigmaLCB(Acquisition):
    def __init__(self, model, best_y, **conf):
        super().__init__(model, **conf)
        self.best_y = best_y
        self.kappa  = conf.get('kappa', 2.0)
        assert(self.model.num_out == 1)

    @property
    def num_obj(self):
        return 2

    @property
    def num_constr(self):
        return 1

    def eval(self, x: Tensor, xe : Tensor) -> Tensor:
        """
        minimize (py, -1 * ps)
        s.t.     LCB  < best_y
        """
        with torch.no_grad():
            out        = torch.zeros(x.shape[0], self.num_obj + self.num_constr)
            py, ps2    = self.model.predict(x, xe)
            noise      = np.sqrt(self.model.noise)
            py        += noise * torch.randn(py.shape)
            ps         = ps2.sqrt()
            lcb        = py - self.kappa * ps
            out[:, 0]  = py.squeeze()
            out[:, 1]  = -1 * ps.squeeze()
            out[:, 2]  = lcb.squeeze() - self.best_y # lcb - best_y < 0
            return out

class MACE(Acquisition):
    def __init__(self, model, best_y, **conf):
        super().__init__(model, **conf)
        self.kappa = conf.get('kappa', 2.0)
        self.eps   = conf.get('eps', 1e-4)
        self.tau   = best_y
    
    @property
    def num_constr(self):
        return 0

    @property
    def num_obj(self):
        return 3

    def eval(self, x : torch.FloatTensor, xe : torch.LongTensor) -> torch.FloatTensor:
        """
        minimize (-1 * EI,  -1 * PI, lcb)
        """
        with torch.no_grad():
            py, ps2   = self.model.predict(x, xe)
            noise     = np.sqrt(2.0) * self.model.noise.sqrt()
            ps        = ps2.sqrt()
            lcb       = (py + noise * torch.randn(py.shape)) - self.kappa * ps
            normed    = ((self.tau - self.eps - py - noise * torch.randn(py.shape)) / ps)
            dist      = Normal(0., 1.)
            log_phi   = dist.log_prob(normed)
            Phi       = dist.cdf(normed)
            PI        = Phi
            EI        = ps * (Phi * normed +  log_phi.exp())
            logEIapp  = ps.log() - 0.5 * normed**2 - (normed**2 - 1).log()
            logPIapp  = -0.5 * normed**2 - torch.log(-1 * normed) - torch.log(torch.sqrt(torch.tensor(2 * np.pi)))

            use_app             = ~((normed > -6) & torch.isfinite(EI.log()) & torch.isfinite(PI.log())).reshape(-1)
            out                 = torch.zeros(x.shape[0], 3)
            out[:, 0]           = lcb.reshape(-1)
            out[:, 1][use_app]  = -1 * logEIapp[use_app].reshape(-1)
            out[:, 2][use_app]  = -1 * logPIapp[use_app].reshape(-1)
            out[:, 1][~use_app] = -1 * EI[~use_app].log().reshape(-1)
            out[:, 2][~use_app] = -1 * PI[~use_app].log().reshape(-1)
            return out

class GeneralAcq(Acquisition):
    def __init__(self, model, num_obj, num_constr, **conf):
        super().__init__(model, **conf)
        self._num_obj    = num_obj
        self._num_constr = num_constr
        self.kappa       = conf.get('kappa', 2.0)
        self.c_kappa     = conf.get('c_kappa', 0.)
        self.use_noise   = conf.get('use_noise', True)
        assert self.model.num_out == self.num_obj + self.num_constr
        assert self.num_obj >= 1

    @property
    def num_obj(self) -> int:
        return self._num_obj

    @property
    def num_constr(self) -> int:
        return self._num_constr

    def eval(self, x : torch.FloatTensor, xe : torch.LongTensor) -> torch.FloatTensor:
        """
        Acquisition function to deal with general constrained, multi-objective optimization problems
        
        Suppose we have $om$ objectives and $cn$ constraints, the problem should has been transformed to :

        Minimize (o1, o1, \dots,  om)
        S.t.     c1 < 0, 
                 c2 < 0, 
                 \dots
                 cb_cn < 0

        In this `GeneralAcq` acquisition function, we calculate lower
        confidence bound of objectives and constraints, and solve the following
        problem:

        Minimize (lcb_o1, lcb_o2, \dots,  lcb_om)
        S.t.     lcb_c1 < 0, 
                 lcb_c2 < 0, 
                 \dots
                 lcb_cn < 0
        """
        with torch.no_grad():
            py, ps2 = self.model.predict(x, xe)
            ps      = ps2.sqrt()
            if self.use_noise:
                noise  = self.model.noise.sqrt()
                py    += noise * torch.randn(py.shape)
            out = torch.ones(py.shape)
            out[:, :self.num_obj] = py[:, :self.num_obj]  - self.kappa   * ps[:, :self.num_obj]
            out[:, self.num_obj:] = py[:, self.num_obj:]  - self.c_kappa * ps[:, self.num_obj:]
        return out



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import pandas as pd
import torch
from torch import Tensor

from .numeric_param     import NumericPara
from .integer_param     import IntegerPara
from .pow_param         import PowPara
from .categorical_param import CategoricalPara
from .bool_param        import BoolPara
from .pow_integer_param         import PowIntegerPara

class DesignSpace:
    def __init__(self):
        self.para_types = {}
        self.register_para_type('num', NumericPara)
        self.register_para_type('pow', PowPara)
        self.register_para_type('pow_int', PowIntegerPara)
        self.register_para_type('int', IntegerPara)
        self.register_para_type('cat', CategoricalPara)
        self.register_para_type('bool', BoolPara)
        self.paras      = {}
        self.para_names = []

    @property
    def numeric_names(self):
        return [p for p in self.para_names if self.paras[p].is_numeric]

    @property
    def enum_names(self):
        return [p for p in self.para_names if self.paras[p].is_categorical]

    @property
    def num_paras(self):
        return len(self.para_names)

    @property
    def num_numeric(self):
        return len(self.numeric_names)

    @property
    def num_categorical(self):
        return len(self.enum_names)

    def parse(self, rec):
        for item in rec:
            assert(item['type'] in self.para_types)
            param = self.para_types[item['type']](item)
            self.paras[param.name] = param
            self.para_names.append(param.name)
        return self

    def register_para_type(self, type_name, para_class):
        """
        User can define their specific parameter type and register the new type
        using this function
        """
        self.para_types[type_name] = para_class

    def sample(self, num_samples = 1):
        """
        df_suggest: suggested initial points
        """
        df = pd.DataFrame(columns = self.para_names)
        for c in df.columns:
            df[c] = self.paras[c].sample(num_samples)
        return df

    def transform(self, data : pd.DataFrame) -> (Tensor, Tensor):
        """
        input: pandas dataframe
        output: xc and xe
        transform data to be within [opt_lb, opt_ub]
        """
        xc = data[self.numeric_names].values.astype(float).copy()
        xe = data[self.enum_names].values.copy()
        for i, name in enumerate(self.numeric_names):
            xc[:, i] = self.paras[name].transform(xc[:, i])
        for i, name in enumerate(self.enum_names):
            xe[:, i] = self.paras[name].transform(xe[:, i])
        return torch.FloatTensor(xc), torch.LongTensor(xe.astype(int))

    def inverse_transform(self, x : Tensor, xe : Tensor) -> pd.DataFrame:
        """
        input: x and xe
        output: pandas dataframe
        """
        with torch.no_grad():
            df_num = pd.DataFrame(columns = self.numeric_names)
            df_cat = pd.DataFrame(columns = self.enum_names)
            for i, name in enumerate(self.numeric_names):
                df_num[name] = self.paras[name].inverse_transform(x.detach().numpy()[:, i])
            for i, name in enumerate(self.enum_names):
                df_cat[name] = self.paras[name].inverse_transform(xe.detach().numpy()[:, i])
            df = pd.concat([df_num, df_cat], axis = 1)
            df = df[self.para_names]
            return df

    @property
    def opt_lb(self):
        lb_numeric = [self.paras[p].opt_lb for p in self.numeric_names]
        lb_enum    = [self.paras[p].opt_lb for p in self.enum_names]
        return torch.tensor(lb_numeric + lb_enum)

    @property
    def opt_ub(self):
        ub_numeric = [self.paras[p].opt_ub for p in self.numeric_names]
        ub_enum    = [self.paras[p].opt_ub for p in self.enum_names]
        return torch.tensor(ub_numeric + ub_enum)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.


import numpy  as np
import pandas as pd
import torch
from torch.quasirandom import SobolEngine
from sklearn.preprocessing import power_transform

from hebo.design_space.design_space import DesignSpace
from hebo.models.model_factory import get_model
from hebo.acquisitions.acq import MACE, Mean, Sigma
from hebo.acq_optimizers.evolution_optimizer import EvolutionOpt

from .abstract_optimizer import AbstractOptimizer

torch.set_num_threads(min(1, torch.get_num_threads()))

class HEBO(AbstractOptimizer):
    support_parallel_opt  = True
    support_combinatorial = True
    support_contextual    = True
    def __init__(self, space, model_name = 'gpy', rand_sample = None, es='nsga2'):
        """
        model_name : surrogate model to be used
        rand_iter  : iterations to perform random sampling
        """
        super().__init__(space)
        self.space       = space
        self.es = es
        self.X           = pd.DataFrame(columns = self.space.para_names)
        self.y           = np.zeros((0, 1))
        self.model_name  = model_name
        self.rand_sample = 1 + self.space.num_paras if rand_sample is None else max(2, rand_sample)
        self.sobol       = SobolEngine(self.space.num_paras, scramble = False)

    def quasi_sample(self, n, fix_input = None): 
        samp    = self.sobol.draw(n)
        samp    = samp * (self.space.opt_ub - self.space.opt_lb) + self.space.opt_lb
        x       = samp[:, :self.space.num_numeric]
        xe      = samp[:, self.space.num_numeric:]
        df_samp = self.space.inverse_transform(x, xe)
        if fix_input is not None:
            for k, v in fix_input.items():
                df_samp[k] = v
        return df_samp

    @property
    def model_config(self):
        if self.model_name == 'gp':
            cfg = {
                    'lr'           : 0.01,
                    'num_epochs'   : 100,
                    'verbose'      : False,
                    'noise_lb'     : 8e-4, 
                    'pred_likeli'  : False
                    }
        elif self.model_name == 'gpy':
            cfg = {
                    'verbose' : False,
                    'warp'    : True,
                    'space'   : self.space
                    }
        elif self.model_name == 'gpy_mlp':
            cfg = {
                    'verbose' : False
                    }
        elif self.model_name == 'rf':
            cfg =  {
                    'n_estimators' : 20
                    }
        else:
            cfg = {}
        if self.space.num_categorical > 0:
            cfg['num_uniqs'] = [len(self.space.paras[name].categories) for name in self.space.enum_names]
        return cfg
            
    def suggest(self, n_suggestions=1, fix_input = None):
        if self.X.shape[0] < self.rand_sample:
            sample = self.quasi_sample(n_suggestions, fix_input)
            return sample
        else:
            X, Xe = self.space.transform(self.X)
            try:
                if self.y.min() <= 0:
                    y = torch.FloatTensor(power_transform(self.y / self.y.std(), method = 'yeo-johnson'))
                else:
                    y = torch.FloatTensor(power_transform(self.y / self.y.std(), method = 'box-cox'))
                    if y.std() < 0.5:
                        y = torch.FloatTensor(power_transform(self.y / self.y.std(), method = 'yeo-johnson'))
                if y.std() < 0.5:
                    raise RuntimeError('Power transformation failed')
                model = get_model(self.model_name, self.space.num_numeric, self.space.num_categorical, 1, **self.model_config)
                model.fit(X, Xe, y)
            except:
                y     = torch.FloatTensor(self.y).clone()
                model = get_model(self.model_name, self.space.num_numeric, self.space.num_categorical, 1, **self.model_config)
                model.fit(X, Xe, y)

            best_id = np.argmin(self.y.squeeze())
            best_x  = self.X.iloc[[best_id]]
            best_y  = y.min()
            py_best, ps2_best = model.predict(*self.space.transform(best_x))
            py_best = py_best.detach().numpy().squeeze()
            ps_best = ps2_best.sqrt().detach().numpy().squeeze()

            iter  = max(1, self.X.shape[0] // n_suggestions)
            upsi  = 0.5
            delta = 0.01
            # kappa = np.sqrt(upsi * 2 * np.log(iter **  (2.0 + self.X.shape[1] / 2.0) * 3 * np.pi**2 / (3 * delta)))
            kappa = np.sqrt(upsi * 2 * ((2.0 + self.X.shape[1] / 2.0) * np.log(iter) + np.log(3 * np.pi**2 / (3 * delta))))

            acq = MACE(model, py_best, kappa = kappa) # LCB < py_best
            mu  = Mean(model)
            sig = Sigma(model, linear_a = -1.)
            opt = EvolutionOpt(self.space, acq, pop = 100, iters = 100, verbose = False, es=self.es)
            rec = opt.optimize(initial_suggest = best_x, fix_input = fix_input).drop_duplicates()
            rec = rec[self.check_unique(rec)]

            cnt = 0
            while rec.shape[0] < n_suggestions:
                rand_rec = self.quasi_sample(n_suggestions - rec.shape[0], fix_input)
                rand_rec = rand_rec[self.check_unique(rand_rec)]
                rec      = rec.append(rand_rec, ignore_index = True)
                cnt +=  1
                if cnt > 3:
                    # sometimes the design space is so small that duplicated sampling is unavoidable
                    break 
            if rec.shape[0] < n_suggestions:
                rand_rec = self.quasi_sample(n_suggestions - rec.shape[0], fix_input)
                rec      = rec.append(rand_rec, ignore_index = True)

            select_id = np.random.choice(rec.shape[0], n_suggestions, replace = False).tolist()
            x_guess   = []
            with torch.no_grad():
                py_all       = mu(*self.space.transform(rec)).squeeze().numpy()
                ps_all       = -1 * sig(*self.space.transform(rec)).squeeze().numpy()
                best_pred_id = np.argmin(py_all)
                best_unce_id = np.argmax(ps_all)
                if best_unce_id not in select_id and n_suggestions > 2:
                    select_id[0]= best_unce_id
                if best_pred_id not in select_id and n_suggestions > 2:
                    select_id[1]= best_pred_id
                rec_selected = rec.iloc[select_id].copy()
            return rec_selected

    def check_unique(self, rec : pd.DataFrame) -> [bool]:
        return (~pd.concat([self.X, rec], axis = 0).duplicated().tail(rec.shape[0]).values).tolist()

    def observe(self, X, y):
        """Feed an observation back.

        Parameters
        ----------
        X : pandas DataFrame
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,1)
            Corresponding values where objective has been evaluated
        """
        valid_id = np.where(np.isfinite(y.reshape(-1)))[0].tolist()
        XX       = X.iloc[valid_id]
        yy       = y[valid_id].reshape(-1, 1)
        self.X   = self.X.append(XX, ignore_index = True)
        self.y   = np.vstack([self.y, yy])



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy  as np
import pandas as pd
import torch

from hebo.design_space.design_space import DesignSpace
from hebo.models.model_factory import get_model
from hebo.acquisitions.acq import LCB
from hebo.acq_optimizers.evolution_optimizer import EvolutionOpt

from .abstract_optimizer import AbstractOptimizer

class BO(AbstractOptimizer):
    support_combinatorial = True
    support_contextual    = True
    def __init__(self, space, model_name = 'gpy', rand_sample = None):
        super().__init__(space)
        self.space       = space
        self.X           = pd.DataFrame(columns = self.space.para_names)
        self.y           = np.zeros((0, 1))
        self.model_name  = model_name
        self.rand_sample = 1 + self.space.num_paras if rand_sample is None else max(2, rand_sample)

    def suggest(self, n_suggestions = 1, fix_input = None):
        assert n_suggestions == 1
        if self.X.shape[0] < self.rand_sample:
            sample = self.space.sample(n_suggestions)
            if fix_input is not None:
                for k, v in fix_input.items():
                    sample[k] = v
            return sample
        else:
            X, Xe     = self.space.transform(self.X)
            y         = torch.FloatTensor(self.y)
            num_uniqs = None if Xe.shape[1] == 0 else [len(self.space.paras[name].categories) for name in self.space.enum_names]
            model     = get_model(self.model_name, X.shape[1], Xe.shape[1], y.shape[1], num_uniqs = num_uniqs, warp = False)
            model.fit(X, Xe, y)

            acq = LCB(model, kappa = 2.0)
            opt = EvolutionOpt(self.space, acq, pop = 100, iters = 100)

            suggest = self.X.iloc[[np.argmin(self.y.reshape(-1))]]
            return opt.optimize(initial_suggest = suggest, fix_input = fix_input)

    def observe(self, X, y):
        """Feed an observation back.

        Parameters
        ----------
        X : pandas DataFrame
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,1)
            Corresponding values where objective has been evaluated
        """
        assert(y.shape[1] == 1)
        valid_id = np.where(np.isfinite(y.reshape(-1)))[0].tolist()
        XX       = X.iloc[valid_id]
        yy       = y[valid_id].reshape(-1, 1)
        self.X   = self.X.append(XX, ignore_index = True)
        self.y   = np.vstack([self.y, yy])



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy  as np
import pandas as pd
import torch

from hebo.design_space.design_space import DesignSpace
from hebo.models.model_factory import get_model, model_dict
from hebo.acquisitions.acq import GeneralAcq
from hebo.acq_optimizers.evolution_optimizer import EvolutionOpt

from .abstract_optimizer import AbstractOptimizer

class GeneralBO(AbstractOptimizer):
    """
    Bayesian optimisation that supports multi-objective and constrained optimization
    """
    def __init__(self,
            space : DesignSpace,
            num_obj:     int   = 1,
            num_constr:  int   = 0,
            rand_sample: int   = None,
            model_name:  str   = 'deep_ensemble',
            model_conf:  dict  = None,
            kappa:       float = 2.,
            c_kappa:     float = 0.,
            use_noise:   bool  = False
            ):
        super().__init__(space)
        self.space       = space
        self.num_obj     = num_obj
        self.num_constr  = num_constr
        self.rand_sample = 1 + self.space.num_paras if rand_sample is None else rand_sample
        self.model_name  = model_name
        self.model_conf  = model_conf if model_conf is not None else {}
        self.X           = pd.DataFrame(columns = self.space.para_names)
        self.y           = np.zeros((0, num_obj + num_constr))
        self.kappa       = kappa
        self.c_kappa     = c_kappa
        self.use_noise   = use_noise
        self.model       = None
        assert model_dict.get(model_name) is not None
        if num_obj + num_constr > 1:
            assert model_dict.get(model_name).support_multi_output

    def suggest(self, n_suggestions = 1, fix_input = None):
        if self.X.shape[0] < self.rand_sample:
            sample = self.space.sample(n_suggestions)
            if fix_input is not None:
                for k, v in fix_input.items():
                    sample[k] = v
            return sample
        else:
            X, Xe      = self.space.transform(self.X)
            y          = torch.FloatTensor(self.y)
            num_uniqs  = None if Xe.shape[1] == 0 else [len(self.space.paras[name].categories) for name in self.space.enum_names]
            self.model = get_model(self.model_name, X.shape[1], Xe.shape[1], y.shape[1], num_uniqs = num_uniqs, **self.model_conf)
            self.model.fit(X, Xe, y)

            acq = GeneralAcq(
                  self.model,
                  self.num_obj,
                  self.num_constr,
                  kappa = self.kappa,
                  c_kappa = self.c_kappa,
                  use_noise = self.use_noise)
            opt     = EvolutionOpt(self.space, acq, pop = 100, iters = 200)
            suggest = opt.optimize()
            with torch.no_grad():
                py, ps2 = self.model.predict(*self.space.transform(suggest))
                largest_uncert_id = np.argmax(np.log(ps2).sum(axis = 1))
            if suggest.shape[0] >= n_suggestions:
                selected_id = np.random.choice(suggest.shape[0], n_suggestions).tolist()
                if largest_uncert_id not in selected_id:
                    selected_id[0] = largest_uncert_id
                return suggest.iloc[selected_id]
            else:
                rand_samp = self.space.sample(n_suggestions - suggest.shape[0])
                suggest   = pd.concat([suggest, rand_samp], axis = 0, ignore_index = True)
                return suggest

    def observe(self, X, y):
        """Feed an observation back.

        Parameters
        ----------
        X : pandas DataFrame
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,1)
            Corresponding values where objective has been evaluated
        """
        valid_id = np.isfinite(y).all(axis = 1)
        XX       = X.iloc[valid_id]
        yy       = y[valid_id]
        self.X   = self.X.append(XX, ignore_index = True)
        self.y   = np.vstack([self.y, yy])
        assert self.y.shape[1] == self.num_obj + self.num_constr

    def select_best(self, rec : pd.DataFrame) -> pd.DataFrame:
        pass

    def get_pf(self, y : np.ndarray) -> pd.DataFrame:
        pass



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
import torch.nn as nn
import torch.nn.functional as F
import gpytorch
import pickle

from torch import Tensor, FloatTensor, LongTensor
from pathlib import Path
from gpytorch.priors  import GammaPrior
from gpytorch.priors.torch_priors import LogNormalPrior
from gpytorch.kernels import ScaleKernel, RBFKernel, MaternKernel, MultitaskKernel
from gpytorch.likelihoods import GaussianLikelihood, MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, ZeroMean, MultitaskMean
from gpytorch.distributions import MultivariateNormal, MultitaskMultivariateNormal
from gpytorch.constraints import Interval, GreaterThan
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from ..util import filter_nan
from ..base_model import BaseModel
from ..layers import EmbTransform
from ..scalers import TorchMinMaxScaler, TorchStandardScaler

class GP(BaseModel):
    support_grad = True
    support_multi_output = True
    def __init__(self, num_cont, num_enum, num_out, **conf):
        super().__init__(num_cont, num_enum, num_out, **conf)
        self.lr           = conf.get('lr', 3e-2)
        self.num_epochs   = conf.get('num_epochs', 100)
        self.verbose = conf.get('verbose', False)
        self.print_every  = conf.get('print_every', 10)
        self.noise_free   = conf.get('noise_free', False)
        self.pred_likeli  = conf.get('pred_likeli', True)
        self.noise_lb     = conf.get('noise_lb', 1e-5)
        self.xscaler      = TorchMinMaxScaler((-1, 1))
        self.yscaler      = TorchStandardScaler()

    def fit_scaler(self, Xc : Tensor, Xe : Tensor, y : Tensor):
        if Xc is not None and Xc.shape[1] > 0:
            self.xscaler.fit(Xc)
        self.yscaler.fit(y)
    
    def xtrans(self, Xc : Tensor, Xe : Tensor, y : Tensor = None):
        if Xc is not None and Xc.shape[1] > 0:
            Xc_t = self.xscaler.transform(Xc)
        else:
            Xc_t = torch.zeros(Xe.shape[0], 0)

        if Xe is None:
            Xe_t = torch.zeros(Xc.shape[0], 0).long()
        else:
            Xe_t = Xe.long()

        if y is not None:
            y_t = self.yscaler.transform(y)
            return Xc_t, Xe_t, y_t
        else:
            return Xc_t, Xe_t

    def fit(self, Xc : Tensor, Xe : Tensor, y : Tensor):
        Xc, Xe, y = filter_nan(Xc, Xe, y, 'all')
        self.fit_scaler(Xc, Xe, y)
        Xc, Xe, y = self.xtrans(Xc, Xe, y)

        assert(Xc.shape[1] == self.num_cont)
        assert(Xe.shape[1] == self.num_enum)
        assert(y.shape[1]  == self.num_out)

        self.Xc = Xc
        self.Xe = Xe
        self.y  = y

        n_constr = GreaterThan(self.noise_lb)
        n_prior  = LogNormalPrior(-4.63, 0.5)
        if self.num_out == 1:
            self.lik = GaussianLikelihood(noise_constraint = n_constr, noise_prior = n_prior)
        else:
            self.lik = MultitaskGaussianLikelihood(num_tasks = self.num_out, noise_constraint = n_constr, noise_prior = n_prior)
        self.gp = GPyTorchModel(self.Xc, self.Xe, self.y, self.lik, **self.conf)

        if self.num_out == 1: # XXX: only tuned for single-output BO
            if self.num_cont > 0:
                self.gp.kern.outputscale = self.y.var()
                lscales = self.gp.kern.base_kernel.lengthscale.detach().clone().view(1, -1)
                for i in range(self.num_cont):
                    lscales[0, i] = torch.pdist(self.Xc[:, i].view(-1, 1)).median().clamp(min = 0.02)
                self.gp.kern.base_kernel.lengthscale = lscales
            if self.noise_free:
                self.gp.likelihood.noise  = self.noise_lb * 1.1
                self.gp.likelihood.raw_noise.requires_grad = False
            else:
                self.gp.likelihood.noise  = max(1e-2, self.noise_lb)

        self.gp.train()
        self.lik.train()

        opt = torch.optim.LBFGS(self.gp.parameters(), lr = self.lr, max_iter = 5, line_search_fn = 'strong_wolfe')
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.lik, self.gp)
        for epoch in range(self.num_epochs):
            def closure():
                dist = self.gp(self.Xc, self.Xe)
                loss = -1 * mll(dist, self.y.squeeze())
                opt.zero_grad()
                loss.backward()
                return loss
            opt.step(closure)
            if self.verbose and ((epoch + 1) % self.print_every == 0 or epoch == 0):
                print('After %d epochs, loss = %g' % (epoch + 1, closure().item()), flush = True)
        self.gp.eval()
        self.lik.eval()

    def predict(self, Xc, Xe):
        Xc, Xe = self.xtrans(Xc, Xe)
        with gpytorch.settings.fast_pred_var(), gpytorch.settings.debug(False):
            pred = self.gp(Xc, Xe)
            if self.pred_likeli:
                pred = self.lik(pred)
            mu_  = pred.mean.reshape(-1, self.num_out)
            var_ = pred.variance.reshape(-1, self.num_out)
        mu  = self.yscaler.inverse_transform(mu_)
        var = var_ * self.yscaler.std**2
        return mu, var

    def sample_y(self, Xc, Xe, n_samples = 1) -> FloatTensor:
        """
        Should return (n_samples, Xc.shape[0], self.num_out) 
        """
        Xc, Xe = self.xtrans(Xc, Xe)
        with gpytorch.settings.debug(False):
            if self.pred_likeli:
                pred = self.lik(self.gp(Xc, Xe))
            else:
                pred = self.gp(Xc, Xe)
            samp = pred.rsample(torch.Size((n_samples,))).view(n_samples, Xc.shape[0], self.num_out)
            return self.yscaler.inverse_transform(samp)

    def sample_f(self):
        raise NotImplementedError('Thompson sampling is not supported for GP, use `sample_y` instead')

    @property
    def noise(self):
        if self.num_out == 1:
            return (self.gp.likelihood.noise * self.yscaler.std**2).view(self.num_out).detach()
        else:
            return (self.gp.likelihood.noise_covar.noise * self.yscaler.std**2).view(self.num_out).detach()

class GPyTorchModel(gpytorch.models.ExactGP):
    def __init__(self, 
            x   : torch.Tensor,
            xe  : torch.Tensor,
            y   : torch.Tensor,
            lik : GaussianLikelihood, 
            **conf):
        super().__init__((x, xe), y.squeeze(), lik)
        mean     = conf.get('mean', ConstantMean())
        kern     = conf.get('kern', ScaleKernel(MaternKernel(nu = 1.5, ard_num_dims = x.shape[1]), outputscale_prior = GammaPrior(0.5, 0.5)))
        kern_emb = conf.get('kern_emb', MaternKernel(nu = 2.5))

        self.multi_task = y.shape[1] > 1
        self.mean  = mean if not self.multi_task else MultitaskMean(mean, num_tasks = y.shape[1])
        if x.shape[1] > 0:
            self.kern = kern if not self.multi_task else MultitaskKernel(kern, num_tasks = y.shape[1])
        if xe.shape[1] > 0:
            assert 'num_uniqs' in conf
            num_uniqs = conf['num_uniqs']
            emb_sizes = conf.get('emb_sizes', None)
            self.emb_trans = EmbTransform(num_uniqs, emb_sizes = emb_sizes)
            self.kern_emb  = kern_emb if not self.multi_task else MultitaskKernel(kern_emb, num_tasks = y.shape[1])

    def forward(self, x, xe):
        m = self.mean(x)
        if x.shape[1] > 0:
            K = self.kern(x)
            if xe.shape[1] > 0:
                x_emb  = self.emb_trans(xe)
                K     *= self.kern_emb(x_emb)
        else:
            K = self.kern_emb(self.emb_trans(xe))
        return MultivariateNormal(m, K) if not self.multi_task else MultitaskMultivariateNormal(m, K)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import numpy  as np
import pandas as pd
import torch
from torch.quasirandom import SobolEngine
from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import get_problem, get_mutation, get_crossover
from pymoo.operators.mixed_variable_operator import MixedVariableMutation, MixedVariableCrossover
from pymoo.optimize import minimize
from pymoo.model.problem import Problem
from pymoo.configuration import Configuration
from ..design_space.design_space import DesignSpace
from ..acquisitions.acq import Acquisition
Configuration.show_compile_hint = False
from pymoo.factory import get_algorithm

class BOProblem(Problem):
    def __init__(self,
            lb    : np.ndarray,
            ub    : np.ndarray,
            acq   : Acquisition,
            space : DesignSpace, 
            fix   : dict = None 
            ):
        super().__init__(len(lb), xl = lb, xu = ub, n_obj = acq.num_obj, n_constr = acq.num_constr)
        self.acq   = acq
        self.space = space
        self.fix   = fix # NOTE: use self.fix to enable contextual BO

    def _evaluate(self, x : np.ndarray, out : dict, *args, **kwargs):
        num_x = x.shape[0]
        xcont = torch.from_numpy(x[:, :self.space.num_numeric].astype(float))
        xenum = torch.from_numpy(x[:, self.space.num_numeric:].astype(float))
        df_x  = self.space.inverse_transform(xcont, xenum)
        if self.fix is not None: # invalidate fixed input, replace with fixed values
            for k, v in self.fix.items():
                df_x[k] = v
        xcont, xenum = self.space.transform(df_x)

        with torch.no_grad():
            acq_eval = self.acq(xcont, xenum).numpy().reshape(num_x, self.acq.num_obj + self.acq.num_constr)
            out['F'] = acq_eval[:, :self.acq.num_obj]

            if self.acq.num_constr > 0:
                out['G'] = acq_eval[:, -1 * self.acq.num_constr:]

class EvolutionOpt:
    def __init__(self,
            design_space : DesignSpace,
            acq          : Acquisition, es,
            **conf):
        self.space      = design_space
        self.es         = es 
        self.acq        = acq
        self.pop        = conf.get('pop', 100)
        self.iter       = conf.get('iters',500)
        self.verbose    = conf.get('verbose', False)
        assert(self.acq.num_obj > 0)

    def get_init_pop(self, initial_suggest : pd.DataFrame = None) -> np.ndarray:
        # init_pop = self.space.sample(self.pop)
        self.eng   = SobolEngine(self.space.num_paras, scramble = True)
        sobol_samp = self.eng.draw(self.pop)
        sobol_samp = sobol_samp * (self.space.opt_ub - self.space.opt_lb) + self.space.opt_lb
        x          = sobol_samp[:, :self.space.num_numeric]
        xe         = sobol_samp[:, self.space.num_numeric:]
        init_pop   = self.space.inverse_transform(x, xe)
        if initial_suggest is not None:
            init_pop = pd.concat([initial_suggest, init_pop], axis = 0).head(self.pop)
        x, xe = self.space.transform(init_pop)
        return np.hstack([x.numpy(), xe.numpy().astype(float)])

    def get_mutation(self):
        mask = []
        for name in (self.space.numeric_names + self.space.enum_names):
            if self.space.paras[name].is_discrete:
                mask.append('int')
            else:
                mask.append('real')

        mutation = MixedVariableMutation(mask, {
            'real' : get_mutation('real_pm', eta = 20), 
            'int'  : get_mutation('int_pm', eta = 20)
        })
        return mutation

    def get_crossover(self):
        mask = []
        for name in (self.space.numeric_names + self.space.enum_names):
            if self.space.paras[name].is_discrete:
                mask.append('int')
            else:
                mask.append('real')

        crossover = MixedVariableCrossover(mask, {
            'real' : get_crossover('real_sbx', eta = 15, prob = 0.9), 
            'int'  : get_crossover('int_sbx', eta = 15, prob = 0.9)
        })
        return crossover

    def optimize(self, initial_suggest : pd.DataFrame = None, fix_input : dict = None) -> pd.DataFrame:
        lb        = self.space.opt_lb.numpy()
        ub        = self.space.opt_ub.numpy()
        prob      = BOProblem(lb, ub, self.acq, self.space, fix_input)
        init_pop  = self.get_init_pop(initial_suggest)
        mutation  = self.get_mutation()
        crossover = self.get_crossover()
        try:
          algo = get_algorithm(self.es, pop_size = self.pop, sampling = init_pop, mutation = mutation, crossover = crossover)
          res   = minimize(prob, algo, ('n_gen', self.iter), verbose = self.verbose)
        except:
          if self.acq.num_obj > 1:
              algo = get_algorithm('nsga2', pop_size = self.pop, sampling = init_pop, mutation = mutation, crossover = crossover)
          else:
              algo = get_algorithm('ga', pop_size = self.pop, sampling = init_pop, mutation = mutation, crossover = crossover)
          res   = minimize(prob, algo, ('n_gen', self.iter), verbose = self.verbose)
        opt_x = res.X.reshape(-1, len(lb)).astype(float)
        
        opt_xcont = torch.from_numpy(opt_x[:, :self.space.num_numeric])
        opt_xenum = torch.from_numpy(opt_x[:, self.space.num_numeric:])
        df_opt    = self.space.inverse_transform(opt_xcont, opt_xenum)
        if fix_input is not None:
            for k, v in fix_input.items():
                df_opt[k] = v
        return df_opt



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import pytest
import torch
import torch.nn as nn
from hebo.models.scalers import TorchStandardScaler, TorchMinMaxScaler, TorchIdentityScaler

@pytest.mark.parametrize('scaler',
        [TorchIdentityScaler(), TorchStandardScaler(), TorchMinMaxScaler((-3, 7))],
        ids = ['Idendity', 'Standard', 'MinMax']
        )
def test_identity(scaler):
    x      = torch.randn(10, 3) * 5 + 4
    scaler = scaler.fit(x)
    xx     = scaler.inverse_transform(scaler.transform(x))
    res    = (x - xx).abs()
    assert(res.sum() < 1e-4)

@pytest.mark.parametrize('scaler',
        [TorchIdentityScaler(), TorchStandardScaler(), TorchMinMaxScaler((-3, 7))],
        ids = ['Idendity', 'Standard', 'MinMax']
        )
def test_identity_same(scaler):
    x      = torch.ones(10, 1) * 5 + 4
    scaler = scaler.fit(x)
    xx     = scaler.inverse_transform(scaler.transform(x))
    res    = (x - xx).abs()
    assert(res.sum() < 1e-4)

    xrand  = torch.randn(10, 3)
    res    = (xrand - scaler.inverse_transform(scaler.transform(xrand))).abs()
    assert(res.sum() < 1e-4)

def test_standard_scaler():
    x      = torch.randn(10, 3) * 5 + 4
    scaler = TorchStandardScaler().fit(x)
    x_tr   = scaler.transform(x)
    assert((x_tr.numpy().mean(axis = 0) - 0).sum() < 1e-6)
    assert((x_tr.numpy().std(axis = 0)  - 1).sum() < 1e-6)

def test_min_max_scaler():
    x      = torch.randn(10, 3) * 5 + 4
    scaler = TorchMinMaxScaler().fit(x)
    x_tr   = scaler.transform(x)
    assert((x_tr.numpy().min(axis = 0) - scaler.range_lb).sum() < 1e-6)
    assert((x_tr.numpy().max(axis = 0) - scaler.range_ub).sum() < 1e-6)

def test_min_max_scaler_same():
    x      = torch.ones(10, 3) * 5 + 4
    scaler = TorchMinMaxScaler().fit(x)
    x_tr   = scaler.transform(x)
    assert((x_tr.numpy().max(axis = 0) - scaler.range_ub).sum() < 1e-6)

@pytest.mark.parametrize('scaler',
        [TorchIdentityScaler(), TorchStandardScaler(), TorchMinMaxScaler((-3, 7))],
        ids = ['Idendity', 'Standard', 'MinMax']
        )
def test_one_sample(scaler):
    x      = torch.zeros(1, 1)
    scaler = scaler.fit(x)
    xx     = scaler.inverse_transform(scaler.transform(x))
    res    = (x - xx).abs()
    assert(res.sum() < 1e-4)

def test_grad():
    x = torch.randn(10, 3) * 5 + 4
    x.requires_grad = True

    scaler0 = TorchStandardScaler().fit(x)
    scaler1 = TorchMinMaxScaler().fit(x)
    scaler2 = TorchIdentityScaler().fit(x)

    model = nn.Sequential(
            scaler0, 
            scaler1, 
            nn.Linear(3, 32), 
            nn.ReLU(), 
            nn.Linear(32, 1))
    y = model(x).sum()
    y.backward()
    assert(x.grad is not None)
    assert(x.grad.shape == x.shape)
    assert(scaler0.mean.grad is None)
    assert(scaler0.std.grad is None)
    assert(scaler1.scale_.grad is None)
    assert(scaler1.min_.grad is None)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
from torch import FloatTensor, LongTensor

def filter_nan(x : FloatTensor, xe : LongTensor, y : FloatTensor, keep_rule = 'any') -> (FloatTensor, LongTensor, FloatTensor):
    assert x  is None or torch.isfinite(x).all()
    assert xe is None or torch.isfinite(xe).all()
    assert torch.isfinite(y).any(), "No valid data in the dataset"

    if keep_rule == 'any':
        valid_id = torch.isfinite(y).any(dim = 1)
    else:
        valid_id = torch.isfinite(y).all(dim = 1)
    x_filtered  = x[valid_id]  if x  is not None else None
    xe_filtered = xe[valid_id] if xe is not None else None
    y_filtered  = y[valid_id]
    return x_filtered, xe_filtered, y_filtered



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys
import numpy as np
from sklearn.preprocessing import MinMaxScaler, StandardScaler
import torch
import torch.nn as nn
import torch.nn.functional as F

class TorchIdentityScaler(nn.Module):
    def __init__(self):
        super().__init__()

    def fit(self, x : torch.FloatTensor):
        return self

    def forward(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return x

    def transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return self.forward(x)

    def inverse_transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return x

class TorchStandardScaler(nn.Module):
    def __init__(self):
        super().__init__()
        self.mean   = None
        self.std    = None
        self.fitted = False

    def fit(self, x : torch.FloatTensor):
        assert(x.dim() == 2)
        with torch.no_grad():
            scaler = StandardScaler()
            scaler.fit(x.detach().numpy())

            self.mean = torch.FloatTensor(scaler.mean_.copy()).view(-1)
            self.std  = torch.FloatTensor(scaler.scale_.copy()).view(-1)
            invalid   = ~(torch.isfinite(self.mean) & torch.isfinite(self.std))
            self.mean[invalid] = 0. # somethime we face data with some all-NaN columns
            self.std[invalid]  = 1.
            return self

    def forward(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return (x - self.mean) / self.std

    def transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return self.forward(x)

    def inverse_transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return x * self.std + self.mean

class TorchMinMaxScaler(nn.Module):
    def __init__(self, range : tuple = (0, 1)):
        super().__init__()
        self.range_lb = float(range[0])
        self.range_ub = float(range[1])
        assert (self.range_ub > self.range_lb )

        self.scale_ = None
        self.min_   = None
        self.fitted = False

    def fit(self, x : torch.FloatTensor):
        assert(x.dim() == 2)
        with torch.no_grad():
            scaler = MinMaxScaler((self.range_lb, self.range_ub))
            scaler.fit(x.detach().numpy())
            self.scale_ = torch.FloatTensor(scaler.scale_)
            self.min_   = torch.FloatTensor(scaler.min_)
            self.fitted = True
        return self

    def forward(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return self.transform(x)

    def transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return self.scale_ * x + self.min_

    def inverse_transform(self, x : torch.FloatTensor) -> torch.FloatTensor:
        return (x - self.min_) / self.scale_



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
import torch.nn as nn
import torch.nn.functional as F

class EmbTransform(nn.Module):
    def __init__(self, num_uniqs, **conf):
        super().__init__()
        self.emb_sizes = conf.get('emb_sizes')
        if self.emb_sizes is None:
            self.emb_sizes = [min(50, 1 + v // 2) for v in num_uniqs]
        
        self.emb = nn.ModuleList([])
        for num_uniq, emb_size in zip(num_uniqs, self.emb_sizes):
            self.emb.append(nn.Embedding(num_uniq, emb_size))
    
    @property
    def num_out(self)->int:
        return sum(self.emb_sizes)

    def forward(self, xe):
        return torch.cat([self.emb[i](xe[:, i]).view(xe.shape[0], -1) for i in range(len(self.emb))], dim = 1)

class OneHotTransform(nn.Module):
    def __init__(self, num_uniqs):
        super().__init__()
        self.num_uniqs = num_uniqs

    @property
    def num_out(self)->int:
        return sum(self.num_uniqs)

    def forward(self, xe):
        return torch.cat([F.one_hot(xe[:, i], self.num_uniqs[i]) for i in range(xe.shape[1])], dim = 1).float()



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys
import numpy as np
from .param import Parameter

class NumericPara(Parameter):
    def __init__(self, param_dict):
        super().__init__(param_dict)
        self.lb   = param_dict['lb']
        self.ub   = param_dict['ub']

    def sample(self, num = 1):
        assert(num > 0)
        return np.random.uniform(self.lb, self.ub, num)

    def transform(self, x):
        return x

    def inverse_transform(self, x):
        return x

    @property
    def is_numeric(self):
        return True

    @property
    def opt_lb(self):
        return self.lb

    @property
    def opt_ub(self):
        return self.ub

    @property
    def is_discrete(self):
        return False



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

from . import evolution_optimizer



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import hebo
import torch
import numpy as p

from hebo.acquisitions.acq import GeneralAcq

class DummyModel:
    def __init__(self, num_obj, num_constr):
        self.num_obj    = num_obj
        self.num_constr = num_constr
        self.num_out    = num_obj + num_constr

    @property
    def noise(self) -> torch.FloatTensor:
        return torch.zeros(self.num_out)

    def predict(self, x, xe):
        num_data = x.shape[0]
        py       = torch.zeros(num_data, self.num_out)
        ps2      = torch.ones(num_data, self.num_out)
        return py, ps2

def test_dummy():
    model = DummyModel(3, 4)
    acq   = GeneralAcq(model, model.num_obj, model.num_constr, kappa = 3, c_kappa = 4, use_noise = True)
    x     = torch.randn(10, 3)
    xe    = torch.ones(10, 3).long()
    acq_v = acq(x, xe)
    assert (acq_v[:, :acq.num_obj] + acq.kappa * torch.ones(10, 3)).abs().max() < 1e-6
    assert (acq_v[:, acq.num_obj:] + acq.c_kappa * torch.ones(10, 4)).abs().max() < 1e-6



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import pandas as pd
import numpy  as np
from abc import ABC, abstractmethod

class Parameter(ABC):
    def __init__(self, param_dict):
        self.param_dict = param_dict
        self.name       = param_dict['name']
        pass

    @abstractmethod
    def sample(self, num = 1) -> pd.DataFrame:
        pass

    @abstractmethod
    def transform(self, x : np.array) -> np.array:
        pass

    @abstractmethod
    def inverse_transform(self, x : np.array) -> np.array:
        pass

    @property
    @abstractmethod
    def is_numeric(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_discrete(self) -> bool:
        """
        Integer and categorical variable
        """
        pass

    @property
    def is_categorical(self) -> bool:
        return not self.is_numeric


    @property
    @abstractmethod
    def opt_lb(self) -> float:
        pass

    @property
    @abstractmethod
    def opt_ub(self) -> float:
        pass




# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
from sklearn.ensemble import RandomForestRegressor
from torch import FloatTensor, LongTensor
import numpy as np

from ..base_model import BaseModel
from ..layers import OneHotTransform
from ..util import filter_nan

class RF(BaseModel):
    def __init__(self, num_cont, num_enum, num_out, **conf):
        super().__init__(num_cont, num_enum, num_out, **conf)
        self.n_estimators = self.conf.get('n_estimators', 100)
        self.rf = RandomForestRegressor(n_estimators = self.n_estimators)
        self.est_noise = torch.zeros(self.num_out)
        if self.num_enum > 0:
            self.one_hot = OneHotTransform(self.conf['num_uniqs'])

    def xtrans(self, Xc : FloatTensor, Xe: LongTensor) -> np.ndarray:
        if self.num_enum == 0:
            return Xc.detach().numpy()
        else:
            Xe_one_hot = self.one_hot(Xe)
            if Xc is None:
                Xc = torch.zeros(Xe.shape[0], 0)
            return torch.cat([Xc, Xe_one_hot], dim = 1).numpy()

    def fit(self, Xc : torch.Tensor, Xe : torch.Tensor, y : torch.Tensor):
        Xc, Xe, y = filter_nan(Xc, Xe, y, 'all')
        Xtr = self.xtrans(Xc, Xe)
        ytr = y.numpy().reshape(-1)
        self.rf.fit(Xtr, ytr)
        mse = np.mean((self.rf.predict(Xtr).reshape(-1) - ytr)**2).reshape(self.num_out)
        self.est_noise = torch.FloatTensor(mse)

    @property
    def noise(self):
        return self.est_noise
    
    def predict(self, Xc : torch.Tensor, Xe : torch.Tensor):
        X     = self.xtrans(Xc, Xe)
        mean  = self.rf.predict(X).reshape(-1, 1)
        preds = []
        for estimator in self.rf.estimators_:
            preds.append(estimator.predict(X).reshape([-1,1]))
        var = np.var(np.concatenate(preds, axis=1), axis=1)
        return torch.FloatTensor(mean.reshape([-1,1])), torch.FloatTensor(var.reshape([-1,1]))




# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.


import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()
    
with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()

setuptools.setup(
        name        = 'HEBO',
        packages    = setuptools.find_packages(),
        description = 'Heteroscedastic Evolutionary Bayesian Optimisation',
        long_description = long_description,
        install_requires = required,
        long_description_content_type = "text/markdown",
        version = "0.1.0",
        author = "Alexander I. Cowen-Rivers & Huawei Noah's Ark",
        license = "MIT",
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
        ],
        url = "https://github.com/Yard1/noah-research/tree/master/HEBO",
        )



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import torch
from torch import Tensor, FloatTensor, LongTensor
from .scalers import TorchStandardScaler, TorchMinMaxScaler
from abc import ABC, abstractmethod

class BaseModel(ABC):
    support_ts           = False
    support_grad         = False
    support_multi_output = False
    support_warm_start   = False

    def __init__(self, 
                 num_cont : int,
                 num_enum : int, 
                 num_out  : int, 
                 **conf):
        """
        Base class for probabilistic regression models
        conf: configuration dict
        """
        self.num_cont = num_cont
        self.num_enum = num_enum
        self.num_out  = num_out
        self.conf     = conf
        assert(self.num_cont >= 0)
        assert(self.num_enum >= 0)
        assert(self.num_out  >  0)
        assert(self.num_cont + self.num_enum > 0)
        if self.num_enum > 0:
            assert 'num_uniqs' in self.conf
            assert type(self.conf['num_uniqs']) == type([])
            assert len(self.conf['num_uniqs'])  == self.num_enum
        if not self.support_multi_output:
            assert self.num_out == 1, "Model only support single-output"

    @abstractmethod
    def fit(self, 
            Xc : FloatTensor,
            Xe : LongTensor,
            y  : FloatTensor):
        pass

    @abstractmethod
    def predict(self, 
                Xc : FloatTensor,
                Xe : LongTensor) -> (FloatTensor, FloatTensor):
        """
        Return (possibly approximated) Gaussian predictive distribution
        Return py and ps2 where py is the mean and ps2 predictive variance.
        """
        pass


    @property
    def noise(self)->FloatTensor:
        """
        Return estimated noise variance, for example, GP can view noise level
        as a hyperparameter and optimize it via MLE, another strategy could be
        using the MSE of training data as noise estimation

        Should return a (self.n_out, ) float tensor
        """
        return torch.zeros(self.num_out)

    def sample_f(self):
        # Thompson sampling
        raise NotImplementedError("Thompson sampling is not supported")

    def sample_y(self, Xc : FloatTensor, Xe : LongTensor, n_samples : int) -> FloatTensor:
        py, ps2 = self.predict(Xc, Xe)
        ps      = ps2.sqrt()
        samp    = torch.zeros(n_samples, py.shape[0], self.num_out)
        for i in range(n_samples):
            samp[i] = py + ps * torch.randn(py.shape)
        return samp



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')

import pytest
from pytest import approx

import numpy as np
import pandas as pd
import torch

from hebo.optimizers.hebo import HEBO
from hebo.optimizers.bo import BO
from hebo.optimizers.general import GeneralBO
from hebo.optimizers.util import parse_space_from_bayesmark

from hebo.design_space.design_space      import DesignSpace
from hebo.design_space.numeric_param     import NumericPara
from hebo.design_space.integer_param     import IntegerPara
from hebo.design_space.pow_param         import PowPara
from hebo.design_space.categorical_param import CategoricalPara
from hebo.design_space.bool_param        import BoolPara

def obj(x : pd.DataFrame) -> np.ndarray:
    return x['x0'].values.astype(float).reshape(-1, 1) ** 2

@pytest.mark.parametrize('model_name', ['gp', 'gpy', 'rf', 'deep_ensemble', 'gpy_mlp']) 
@pytest.mark.parametrize('opt_cls', [BO, HEBO, GeneralBO], ids = ['bo', 'hebo', 'general'])
def test_opt(model_name, opt_cls):
    space = DesignSpace().parse([
        {'name' : 'x0', 'type' : 'num', 'lb' : -3, 'ub' : 7},
        {'name' : 'x1', 'type' : 'cat', 'categories' : ['a', 'b', 'c']}
        ])
    opt = opt_cls(space, rand_sample = 10, model_name = model_name)
    for i in range(11):
        num_suggest = 8 if opt.support_parallel_opt else 1
        rec = opt.suggest(n_suggestions = num_suggest)
        y   = obj(rec)
        if y.shape[0] > 1 and i > 0:
            y[np.argmax(y.reshape(-1))] = np.inf
        opt.observe(rec, y)
        if opt.y.shape[0] > 11:
            break

@pytest.mark.parametrize('opt_cls', [BO, HEBO, GeneralBO], ids = ['bo', 'hebo', 'general'])
def test_contextual_opt(opt_cls):
    space = DesignSpace().parse([
        {'name' : 'x0', 'type' : 'int', 'lb' : 0, 'ub' : 7},
        {'name' : 'x1', 'type' : 'int', 'lb' : 0, 'ub' : 7},
        ])
    opt = opt_cls(space, rand_sample = 2, model_name = 'rf')
    for i in range(2):
        n_suggestions = 8 if opt.support_parallel_opt else 1
        rec = opt.suggest(n_suggestions = n_suggestions, fix_input = {'x0' : 3})
        y   = (rec[['x0', 'x1']].values ** 2).sum(axis = 1, keepdims =True)
        assert((rec['x0'] == 3).all())
        opt.observe(rec, y)

def test_bayesmark_parser():
    api_config = \
            {'hidden_layer_sizes': {'type': 'int', 'space': 'linear', 'range': (50, 200)},
                    'alpha': {'type': 'real', 'space': 'log', 'range': (1e-5, 1e1)},
                    'batch_size': {'type': 'int', 'space': 'linear', 'range': (10, 250)},
                    'learning_rate_init': {'type': 'real', 'space': 'log', 'range': (1e-5, 1e-1)},
                    'tol': {'type': 'real', 'space': 'log', 'range': (1e-5, 1e-1)},
                    'validation_fraction': {'type': 'real', 'space': 'logit', 'range': (0.1, 0.9)},
                    'beta_1': {'type': 'real', 'space': 'logit', 'range': (0.5, 0.99)},
                    'beta_2': {'type': 'real', 'space': 'logit', 'range': (0.9, 1.0 - 1e-6)},
                    'epsilon': {'type': 'real', 'space': 'log', 'range': (1e-9, 1e-6)}, 
                    'activation'  : {'type' : 'cat', 'values' : ['sigmoid', 'tanh', 'relu']}, 
                    'use_dropout'  : {'type' : 'bool'}, 
                    'dummy': {'type': 'real', 'space': 'linear', 'range': (4.1, 9.2)}}
    space = parse_space_from_bayesmark(api_config)
    assert(isinstance(space.paras['hidden_layer_sizes'], IntegerPara))
    assert(isinstance(space.paras['alpha'], PowPara))
    assert(isinstance(space.paras['batch_size'], IntegerPara))
    assert(isinstance(space.paras['learning_rate_init'], PowPara))
    assert(isinstance(space.paras['tol'], PowPara))
    assert(isinstance(space.paras['validation_fraction'], PowPara))
    assert(isinstance(space.paras['beta_1'], PowPara))
    assert(isinstance(space.paras['beta_2'], PowPara))
    assert(isinstance(space.paras['epsilon'], PowPara))
    assert(isinstance(space.paras['activation'], CategoricalPara))
    assert(isinstance(space.paras['use_dropout'], BoolPara))

    with pytest.raises(AssertionError):
        api_config = {'hidden_layer_sizes': {'type': '_int', 'space': 'linear', 'range': (50, 200)}}
        space      = parse_space_from_bayesmark(api_config)



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys, os
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + '/../')
import torch
import torch.nn as nn
import pytest

from hebo.models.layers import OneHotTransform, EmbTransform

def test_encoders():
    layer1 = EmbTransform([5, 5], emb_sizes = [1, 1])
    layer2 = EmbTransform([5, 5], emb_sizes = [2, 2])
    layer3 = OneHotTransform([5, 5])
    layer4 = EmbTransform([5, 5]) 

    xe     = torch.randint(5, (10, 2))
    assert(layer1(xe).shape[1] == 2)
    assert(layer1.num_out == 2)

    assert(layer2(xe).shape[1] == 4)
    assert(layer2.num_out == 4)

    assert(layer4(xe).shape[1] == layer4.num_out)

    assert(layer3(xe).shape[1] == 10)
    assert(layer3.num_out == 10)

    assert(torch.all((layer3(xe) == 0) | (layer3(xe) == 1)))
    assert(torch.all(layer3(xe).sum(dim = 1) == xe.shape[1]))

    model1 = nn.Sequential(
            OneHotTransform([5, 5]), 
            nn.Linear(10, 1))
    model2 = nn.Sequential(
            EmbTransform([5, 5], emb_sizes = [2, 2]), 
            nn.Linear(4, 1))
    assert(model1(xe).shape == model2(xe).shape)

    with pytest.raises((RuntimeError, IndexError)):
        xe = torch.randint(50, (100, 2))
        layer1(xe)

    with pytest.raises(RuntimeError):
        xe = torch.randint(50, (100, 2))
        layer3(xe)
